------------------------------------ MODULE DoRiS ---------------------------------------
(***************************************************************************************)
(* Naturals, Reals, Sequences and TLC are TLA+ modules (libraries)                     *)
(* imported by the specification.                                                      *)
(***************************************************************************************)
EXTENDS Naturals, Reals, Sequences, TLC

(***************************************************************************************)
(* VARIABLES                                                                           *)
(*                                                                                     *)
(* Shared: distributed global state variable shared by the servers made up             *)
(* of six fields:                                                                      *)
(* - soft: the group membership of the soft ring.                                      *)
(* - chipTimer: an increasing and cyclic timer that range from 0 to deltaChip.         *)
(* - chipCount: an increasing and cyclic counter that range from 0 to nServ.           *)
(* - cycleCount: an increasing and cyclic counter that range from 0 to horizon.        *)
(* - medium: If no message is being transmitted,  medium equals {}.                    *)
(*   Else, medium stores the message being transmitted.                                *)
(* - macTimer: a decreasing timer that specifies the time remaining in                 *)
(*   order to complete an on-going message transmission.                               *)
(*   As a special case, macTimer = 0 => medium = {}                                    *)
(*                                                                                     *)
(* HardState: local state variable (tuple) associated to the hard ring made up         *)
(* of four fields:                                                                     *)
(* - msg: list of hard messages stored in local buffers after their                    *)
(*   reception by the network device.                                                  *)
(* - execTimer: decreasing timer that specifies the time remaining                     *)
(*   in order to complete the processing of a hard message to nServ.                   *)
(* - res: reservation list for the nServ next chips.                                   *)
(* - cons: counter that represents the number of elementary messages                   *)  
(*   received in a complete DoRiS cycle.                                               *) 
(*                                                                                     *)
(* SoftState: local state variable (tuple) associated to the soft ring                 *)
(* made up of three fields:                                                            *)
(* - token: counter used to rule the token circulation during a soft window.           *)
(* - list: list of soft messages waiting to be transmitted.                            *)
(* - count: the number of soft messages received during a soft window .                *)
(*                                                                                     *)
(* History: an observer global state variable made up of two fields,                   *)
(* used to check specific temporal properties:                                         *)
(* - elem:  the number of elementary messages sent in a cycle.                         *)
(* - rese:  the number of reservation messages sent in a cycle.                        *)
(***************************************************************************************)
VARIABLES Shared, HardState, SoftState, History

(***************************************************************************************)
(* CONSTANTS (Their values are defined in the file DoRiS.cfg)                          *)
(*                                                                                     *)
(* - nServ: number of DoRiS servers.                                                   *)
(* - deltaChip: duration of a chip.                                                    *)
(* - delta: transmission time of a hard message (smallest Ethernet frame - 64 bytes).  *)
(* - pi: processing time of a hard message by the slowest device.                      *)
(* - maxTxTime: transmission time of the largest Ethernet frame (1500 bytes).          *)
(* - horizon: Number of total cycles used to check a DoRiS model.                      *)
(***************************************************************************************)
CONSTANTS nServ, deltaChip, delta, pi, maxTxTime, horizon

(***************************************************************************************)
(* Miscellaneous definitions                                                           *)
(* - min(S): function that returns the minimum value of S.                             *)
(* - next(i, S): function that returns the clothest value from i in ring S.            *)
(* - ServID: set of servers identifiers.                                               *)
(* - HardServ and SoftServ: server local threads dealing with the hard and             *)
(* soft communication.                                                                 *)
(* - HardRing: hard ring.                                                              *)
(* - SoftRing(S): function that, given a set S of identifiers, returns the soft ring   *)
(*    membership.                                                                      *)
(* - hardID(s): function that, given a hard server thread s, returns the identifier    *)
(*   of that thread.                                                                   *)
(* - softID(s, S): function that, given a soft server thread s and a soft              *)
(*   membership S, returns the identifier of that thread.                              *)
(* - vars: the set of all variables of the specification                               *)
(***************************************************************************************)
min(S)  == CHOOSE m \in S : \A y \in S : m =< y 

next(i, R) ==
    IF \A j \in R: j \leq i THEN min(R) ELSE min( { j \in R : j > i } )
 
ServID == 1..nServ                                 
HardServ == [ i \in ServID |-> << "Hard", i >> ]
SoftServ == [ j \in ServID |-> << "Soft", j >> ]
HardRing == { HardServ[i]: i \in ServID }
SoftRing(S) == { SoftServ[j]: j \in S }

hardID(s) == CHOOSE i \in ServID : HardServ[i] = s 
softID(s, S) == CHOOSE j \in S : SoftServ[j] = s 
          
vars == << Shared, SoftState, HardState, History >>           
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* Four alternative soft message list fabric                                           *)
(* (prefixed definition are not used when model-checking).                             *)
(***************************************************************************************)

list(j, cycle) ==
    CASE  j \in { cycle } -> [ i \in 1..4 |-> [ txTime |-> maxTxTime ] ]
      []  j \in { 2, 3 } \ { cycle } -> << >>
      []  j \in { 4, 7 } \ { cycle } -> [ i \in 1..2 |-> [ txTime |-> maxTxTime ] ]
      []  OTHER -> 1 :> [ txTime |-> maxTxTime ]

A_list(j) == 1 :> [ txTime |-> maxTxTime ]

B_list(j) == IF j = 1
    THEN 1 :> [ txTime |-> maxTxTime ] @@ 2 :> [ txTime |-> maxTxTime ] @@
         3 :> [ txTime |-> maxTxTime ]
    ELSE IF j = 2 THEN << >> ELSE 1 :> [ txTime |-> maxTxTime ] 

C_list(j, cycle) ==
    CASE  j \in { cycle } -> [ i \in 1..4 |-> [ txTime |-> maxTxTime ] ]
      []  OTHER -> << >>
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* Initializations of the variables.                                                   *)
(* The Shared variable represents the common knowledge of each server.                 *)
(***************************************************************************************)

Init ==
    /\ Shared = [ soft |-> 1..nServ,
                  chipTimer |-> 0,
                  chipCount |-> 1,
                  cycleCount |-> 1,
                  macTimer |-> 0,
                  medium |-> {} ]
                  
    /\ HardState = [ i \in ServID |-> [ msg |-> << >>, 
                                        execTimer |-> Infinity, 
                                        res |-> [ j \in ServID |-> -1 ],
                                        cons |-> [ j \in ServID |-> 1 ] ] ]
                                        
    /\ SoftState = [ j \in Shared.soft |-> [ token |-> 1,
                                             list |-> list(j, 1),
                                             count |-> 0 ] ] @@
                   [ j \in ServID \ Shared.soft |-> [ token |-> -1,
                                                      list |-> list(j, 1),
                                                      count |-> 0 ] ]
                                                      
    /\ History = [ elem |-> 0,
                   rese |-> 0 ]
-----------------------------------------------------------------------------------------
(*                      `^ \newline ^'                                                 *)
(***************************************************************************************)
(* The elementary sending action.                                                      *)
(* reservation(i) is an arbitrary reservation function                                 *)
(* that states that task "i" wants to reserve as much slots as possible.               *)
(***************************************************************************************)

reservation(i) ==
    IF \A j \in ServID : HardState[i].cons[j] = 1
    THEN { j \in ServID: HardState[i].res[j] = -1 }
    ELSE { ( ( (i - 1) + (nServ - 1) )  %  nServ ) + 1 }

ElemSend(s) ==
    /\ Shared.medium = {}
    /\ Shared.chipTimer = 0
    /\ LET i    ==  hardID(s)
           flag ==  IF SoftState[i].list # << >> THEN 1 ELSE 0
       IN /\ Shared.chipCount = i
          /\ LET resSet == reservation(i) 
             IN /\ Shared' = [ Shared EXCEPT
                       !.macTimer = delta,
                       !.medium = { [ id |-> i, type |-> "hard",
                                      res |-> resSet, softFlag |-> flag ] } ]
                /\ HardState' = [ HardState EXCEPT 
                       ![i].cons[i] = 1,
                       ![i].res = [ j \in ServID |-> IF j \in resSet THEN i ELSE @[j] ] ]
                /\ SoftState' = [ SoftState EXCEPT
                       ![i].token = IF flag = 0 THEN -1 ELSE @ ]
                /\ History' = [ History EXCEPT !.elem = @ + 1 ]
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* The reservation sending action.                                                     *)
(***************************************************************************************)

ReseSend(s) == Shared.medium = {}
    /\ Shared.chipTimer = delta
    /\ LET i == hardID(s)
       IN /\ HardState[i].res[Shared.chipCount] = i
          /\ Shared' = [ Shared EXCEPT 
                 !.macTimer = delta, 
                 !.medium = { [ id |-> i, type |-> "hard", res |-> {-1} ] } ]
          /\ HardState' = [ j \in ServID |-> [ HardState[j] EXCEPT 
                 !.res[Shared.chipCount] = -1 ] ]
          /\ History' = [ History EXCEPT !.rese = @ + 1 ]
          /\ UNCHANGED SoftState       
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* The soft window action.                                                             *)
(* "Failed" specifies the set of omission failures that take place as a function of    *)
(* "Shared.chipCount".                                                                 *)
(* tailList(msgList): function that remove the first message of msgList                *)
(* (if msgList is not empty).                                                          *)
(* lenMsg(i): function that returns the time needed to transmit the first message      *)
(* of server "i" list of pending soft messages, if this list is not empty. Else, it    *)
(* returns "delta".                                                                      *)
(***************************************************************************************)

Failed == CASE  Shared.chipCount = 2 -> { 3 }
	        []  Shared.chipCount \in {3, 4} -> { 3, 5 }
	        []  Shared.chipCount = 5 -> { 3, 5 }
	        []  OTHER -> { }

A_Failed == {}

B_Failed == LET j == CHOOSE i \in ServID : TRUE IN { j }

tailList(msgList) ==
	IF msgList = << >> THEN msgList ELSE Tail(msgList)
	 
lenMsg(i) ==
    IF SoftState[i].list # << >> THEN Head(SoftState[i].list).txTime ELSE delta

SoftSend(s) == 
    /\ Shared.medium = {}
    /\ 2 * delta \leq Shared.chipTimer /\ Shared.chipTimer \leq deltaChip
    /\ LET i      == softID(s, Shared.soft)
           lenTX  == lenMsg(i)
           d      == Shared.chipTimer + lenTX
           consis == \A j \in ServID : HardState[i].cons[j] = 1
           wait   == (d > deltaChip) \/  (\neg consis) 
           noMsg  == (i \in Failed) \/ wait 
       IN /\ i = SoftState[i].token
          /\ Shared' = [ Shared EXCEPT
                !.macTimer = IF noMsg THEN Infinity ELSE lenTX,
                !.medium = IF noMsg  THEN @ ELSE { [ id |-> i, type |-> "soft" ] } ]
          /\ SoftState' = [ SoftState EXCEPT 
                ![i].list = IF  wait THEN @ ELSE tailList(@),
                ![i].token = CASE wait -> @ 
                               [] \neg consis -> -1 
                               [] OTHER -> next(i, Shared.soft),
                ![i].count = IF noMsg  THEN @ ELSE @ + 1 ]
          /\ UNCHANGED << HardState, History >>
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* The hard and soft messages receiving actions.                                       *)
(* The hard and soft messages receiving actions.                                       *)
(* NoRecvSet(m): function that defines omission failures of the "m" reception action.  *)
(* tokenUpdate(m): function that updates the "token" counter.                          *)
(***************************************************************************************)

NoRecvSet(m) == IF Shared.chipCount \in { 2, 4 } THEN { m.id, 2 } ELSE { m.id }

C_NoRecvSet(m) == IF Shared.chipCount \in { 2, 4 } THEN { m.id, 3 } ELSE { m.id }

A_NoRecvSet(m) == LET i == CHOOSE j \in ServID : TRUE IN { m.id, i }

B_NoRecvSet(m) == {m.id}

tokenUpdate(m) == SoftState' = 
     [ i \in (ServID \ Shared.soft) \cup NoRecvSet(m) |-> SoftState[i] ] @@
     [ i \in Shared.soft \ NoRecvSet(m) |-> [ SoftState[i] EXCEPT
               !.token = IF @ = m.id THEN next(m.id, Shared'.soft) ELSE @ ] ]

HardRecv(m) ==
    /\ CASE  m.res # {-1} /\ m.softFlag = 1 
         ->  /\ Shared' = [ Shared EXCEPT !.medium = {}, !.soft = @ \cup {m.id} ]
             /\ UNCHANGED SoftState
         []  m.res # {-1} /\ m.softFlag = 0
         ->  /\ Shared' = [ Shared EXCEPT !.medium = {}, !.soft = @ \ {m.id} ]
             /\ tokenUpdate(m)
         []  m.res = {-1}
         ->  /\ Shared' = [ Shared EXCEPT !.medium = {} ]
             /\ UNCHANGED SoftState
    /\ HardState' =
          [ i \in NoRecvSet(m) |-> HardState[i] ] @@
          [ i \in ServID \ NoRecvSet(m) |-> [ HardState[i] EXCEPT
               !.execTimer = IF Len( HardState[i].msg ) = 0 THEN pi ELSE @,
               !.cons[m.id] = IF m.res # {-1} THEN 1 ELSE @,            
               !.msg = Append( @, m), !.res =
                     IF m.res = {-1}
                     THEN [ j \in ServID |-> IF j = m.id THEN -1 ELSE @[j] ]
                     ELSE [ j \in ServID |-> IF j \in m.res THEN m.id ELSE @[j] ] ] ]
    /\ UNCHANGED History

SoftRecv(m) ==
    /\ LET lastSoft == /\ 2 * delta \leq Shared.chipTimer
                       /\ Shared.chipTimer \leq deltaChip
                       /\ \A j \in ServID : SoftState[j].token = -1
       IN Shared' = [ Shared EXCEPT
                          !.medium = {},
                          !.macTimer = IF lastSoft THEN Infinity ELSE @,
                          !.soft = IF lastSoft THEN {} ELSE @ ]
    /\ SoftState' = [ j \in ( ServID \ Shared.soft ) \cup {m.id} |-> SoftState[j] ] @@
                    [ j \in Shared.soft \ {m.id} |-> [ SoftState[j] EXCEPT
                          !.token = next(@ ,Shared.soft),
                          !.count = @ + 1 ] ]
    /\ UNCHANGED << HardState, History >>
      
Receive(m) ==
    /\ Shared.macTimer = 0
    /\ CASE  m.type = "hard" -> HardRecv( m )
         []  m.type = "soft" -> SoftRecv( m ) 
      
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* NextTick is the only actions that update timers according to passage of time.       *)
(* timerUpdate(): function that updates all timers                                     *)
(***************************************************************************************)

timerUpdate(d, noRese, noSoft) ==
    /\ Shared' = [ Shared EXCEPT
           !.chipTimer = @ + d,
           !.macTimer = CASE  noRese -> @  []  noSoft -> Infinity
                          []  @ = Infinity -> Infinity  []  OTHER -> @ - d ]
    /\ HardState' = [ i \in ServID |-> [ HardState[i] EXCEPT
           !.msg = IF HardState[i].execTimer - d = 0  THEN Tail( @ ) ELSE @,
           !.execTimer = IF @ - d = 0
                         THEN IF Len( HardState[i].msg ) > 1 THEN pi ELSE Infinity
                         ELSE IF @ = Infinity THEN @ ELSE @ - d ] ]

minIncVal(tmp, noRese, noSoft) ==
    CASE  noRese  -> min( { delta } \cup tmp )
      []  noSoft  -> min( tmp )
      []  OTHER   -> min( { Shared.macTimer } \cup tmp )

NextTick ==
    LET noRese == /\ Shared.medium = {}
                  /\ Shared.chipTimer = delta
                  /\ \A i \in ServID : HardState[i].res[Shared.chipCount] # i
        noSoft == /\ 2 * delta \leq Shared.chipTimer
                  /\ Shared.chipTimer \leq deltaChip
                  /\ Shared.medium = {}
                  /\ \A j \in Shared.soft : SoftState[j].token # j
        tmp == { HardState[i].execTimer : i \in ServID } \cup 
               { deltaChip - Shared.chipTimer }
        d == minIncVal(tmp, noRese, noSoft) 
    IN /\ d > 0
       /\ timerUpdate(d, noRese, noSoft) 
       /\ UNCHANGED << SoftState, History >>              

-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* NextChip increment ChipCount modulo the number of task.                             *)
(* It sets the arbitrary soft sequence of message of each process during every cycle.  *)
(* Alltogether, NextTick and NextChip implements the time circularity of the model.    *)
(* The "count" counter allows to check if a soft message has been sent in the previous *)
(* chip. If it is not the case, all processes increment the token value by 1 at        *)
(* the end of the chip.                                                                *)
(***************************************************************************************)         

NextChip ==
    /\ Shared.medium = {}
    /\ Shared.chipTimer = deltaChip
    /\ LET Overflow == \E j \in Shared.soft : Len( SoftState[j].list ) > 14
           TimeCircle == Shared.cycleCount = horizon
           NextCycle == Shared.chipCount = nServ
       IN /\ Shared' = [ Shared EXCEPT !.macTimer = 0, 
               !.chipCount = (@  %  nServ ) + 1,
               !.chipTimer = IF Overflow THEN -1 ELSE 0,
               !.cycleCount = IF TimeCircle THEN 1 ELSE IF NextCycle THEN @ + 1 ELSE @ ]
          /\ SoftState' = [ j \in (ServID \ Shared.soft) |-> SoftState[j] ] @@ 
                           [ j \in Shared.soft |-> [ SoftState[j] EXCEPT !.count = 0,
               !.token = IF SoftState[j].count = 0 THEN  next(@, Shared.soft) ELSE @,
               !.list = IF NextCycle THEN @ \o list(j, Shared.cycleCount) ELSE @ ] ]
          /\ HardState' = [ i \in ServID |-> [ HardState[i] EXCEPT
                                                   !.cons[Shared'.chipCount] = 0 ] ]
          /\ IF NextCycle
             THEN History' = [ elem |-> 0, rese |-> 0 ]
             ELSE UNCHANGED History
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* Next is the disjunct of steps choices that doesn't alterate the timers              *)
(***************************************************************************************)
Next  ==  \/ \E s \in HardRing : ElemSend(s) \/ ReseSend(s)
          \/ \E t \in SoftRing(Shared.soft) : SoftSend(t)
          \/ \E msg \in Shared.medium : Receive(msg)
     
(***************************************************************************************)
(* Tick defines the passage of time.                                                   *)
(***************************************************************************************)
Tick  == NextTick \/ NextChip

(***************************************************************************************)
(* Liveness is a condition that warrants the eventual flow of time.                    *)
(***************************************************************************************)
Liveness ==  []<> Tick            
 
(***************************************************************************************)
(* The main formula of the specification                                               *)
(***************************************************************************************)
Spec  ==  Init /\ [] [ Next  \/ Tick ]_vars /\ Liveness 
-----------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------
(***************************************************************************************)
(* The following temporal properties are checked when specified in the DoRiS.cfg       *)
(* configuration file.                                                                 *)
(* "TypeInvariance" checks the variables type invariance.                              *)
(* "CollisionAvoidance" checks that the token can be hold by only one task or process  *)
(* at once.                                                                            *)
(* "HardRingCorrectness" makes use of the "History" variable to check that all         *)
(* mandatory elementary message are sent in a DoRiS cycle.                             *)
(* "ReservationSafety" guaranty that if a task holds the token in a reservation slot,  *)
(* then all other tasks are aware of its reservation.                                  *)
(* "SoftRingFairness" guaranty that each process will eventually receive the token.    *)
(* "Omission" states that some task omission failure takes place.                      *)
(* "Failure" states that some process failure takes place.                             *)
(* "NoReservationSafety", "NoCollisionAvoidance" and "NoOmission" are contradiction    *)
(* of the respective properties, used to generate handfull conter-examples.            *)
(***************************************************************************************)

HardMsg == Seq( [ id: ServID, type: {"hard"}, res : { {-1} } ] \cup 
     [ id: ServID, type: {"hard"}, res: SUBSET( {-1} \cup ServID ), softFlag : 0..1 ] )            

MediumMsg == { m: m \in [ id: ServID, type: {"soft"} ] \cup
     [ id: ServID, type: {"hard"}, res : SUBSET( {-1} ) ] \cup
     [ id: ServID, type: {"hard"}, res: SUBSET( {-1} \cup ServID ), softFlag : 0..1  ] }

TypeInvariance ==
    /\ Shared.chipCount \in 1..nServ
    /\ Shared.cycleCount \in 1..horizon
    /\ Shared.chipTimer \in 0..deltaChip
    /\ Shared.macTimer \in 0..maxTxTime \cup {Infinity}
    /\ \A m \in Shared.medium : m \in MediumMsg
    /\ SoftState \in [ ServID ->
           [ token : ServID \cup {-1},
             count : 0..50,
             list : {<< >>} \cup Seq( [ txTime : { maxTxTime } ] ) ] ]
    /\ HardState \in [ ServID ->
           [ msg : {<< >>} \cup HardMsg,
             res : [ ServID -> {-1} \cup ServID  ], 
             execTimer : 0..pi \cup {Infinity},
             cons : [ ServID -> 0..1 ]  ] ]

Send(s) ==  \/ /\ s \in HardRing 
               /\ \/ ENABLED ElemSend(s)
                  \/ ENABLED ReseSend(s)
            \/ /\ s \in SoftRing(Shared.soft) 
               /\ ENABLED SoftSend(s)

CollisionAvoidance == 
    \A s, t \in HardRing \cup SoftRing(ServID):
          [] ( ENABLED ( Send(s) /\ Send(t) ) => ( s = t ) ) 

NoCollisionAvoidance == 
    \E s, t \in HardRing \cup SoftRing(ServID):
          <> ( ( s # t ) /\ ENABLED ( Send(s) /\ Send(t) ) )

HardRingCorrectness ==
    /\ \A s \in HardRing : [] ( Len( HardState[hardID(s)].msg ) \leq 3 )
    /\ [] ( ENABLED NextChip => History.elem = Shared.chipCount )
   
ReservationSafety == 
    [] \A chip, j \in ServID : /\ ENABLED ReseSend(HardServ[j])
                               /\ Shared.chipCount = chip
         =>  /\ HardState[j].res[chip] = j 
             /\ \A i \in ServID\{j}: HardState[i].res[chip] \in { j, -1 }
                 
SoftRingFairness ==
    /\ \A i \in ServID : [] ( i \in Shared.soft 
            => ( SoftState[i].list # << >> => <> ( i = SoftState[i].token ) ) )
    /\ []<> ( \A i \in ServID \ Failed : i \in Shared.soft => Len(SoftState[i].list) = 0 )

NoReservationSafety ==
    []<>  \E chip \in ServID: \E j \in ServID:
             /\ HardState[j].res[chip] # -1
             /\ ENABLED ReseSend(HardServ[j])
             /\ Shared.chipCount = chip
             /\ \E i \in ServID\{j}: \neg HardState[i].res[chip] \in { j, -1 }

Omission == \E s \in HardRing :
    <> ( ENABLED ElemSend(s)  /\ \E i \in ServID : HardState[hardID(s)].cons[i] = 0 )

NoOmission == \A s \in HardRing :
    [] ( ENABLED ElemSend(s)  => (\A i \in ServID : HardState[hardID(s)].cons = 1 ) )
         
Failure == []<> ( \E p \in SoftRing(ServID) : softID(p, ServID) \in Failed )
=========================================================================================
